#!/usr/bin/env node

import puppeteer from 'puppeteer-extra'
import StealthPlugin from 'puppeteer-extra-plugin-stealth'
import sleep from 'sleep-promise'
import { main } from '../lib/main.mjs'
import logger from '../lib/logger.mjs'
import yargs from 'yargs'
import { hideBin } from 'yargs/helpers'
import fs from 'fs/promises'
import yaml from 'js-yaml'

puppeteer.use(StealthPlugin())

const argv = yargs(hideBin(process.argv))
  .scriptName('oroshi')
  .command('$0 <file>', 'run scenario file')
  .option('endpoint', {
    describe: 'connect to a puppeteer instance via its ws endpoint',
    type: 'string',
  })
  .option('debug', {
    describe: 'prefer headful browser and output debug logging',
    type: 'boolean',
  })
  .help()
  .argv

let browser

if (argv.debug) {
  logger.level = 'debug'
}

async function connect(endpoint, debug) {
  while (true) {
    try {
      if (endpoint) {
        logger.info('Connecting to puppeteer host')
        browser = await puppeteer.connect({
          browserWSEndpoint: endpoint,
        })
      } else {
        logger.info('Launching puppeteer instance')
        browser = await puppeteer.launch({
          headless: !debug,
          slowMo: debug,
        })
      }

      return browser
    } catch (err) {
      logger.error(err)
      logger.info('Retry connection in 10 sec...')
      await sleep(10000)
    }
  }
}

async function parseScenario(filePath) {
  try {
    const scenario = yaml.load(await fs.readFile(filePath, 'utf8'))
    return scenario
  } catch (err) {
    logger.error(err)
  }
}

async function run(argv) {
  let scenario

  try {
    await fs.stat(argv.file)
    scenario = await parseScenario(argv.file)
  } catch (err) {
    logger.error('Missing scenario file')
    process.exit(1)
  }

  try {
    if (!browser || !browser.isConnected()) {
      browser = await connect(argv.endpoint, argv.debug)
    }

    logger.info(`Scraping...`)
    await main(browser, scenario)
  } catch (err) {
    logger.error(err)
  } finally {
    if (argv.endpoint) {
      await browser.disconnect()
    } else {
      await browser.close()
    }
  }
}

run(argv)
