import logger from './logger.mjs'
import sleep from 'sleep-promise'
import { sendWebhook } from './notifications.mjs'
import parseDuration from 'parse-duration'

const scannedUrl = new Set()

export async function main(browser, scenario) {
  await runScenario(scenario, browser)
}

async function runScenario(scenario, browser) {
  for (let pageScenario of scenario.pages) {
    let page = await browser.newPage()

    await runPageScenario(pageScenario, page)

    await page.close()
  }
}

async function runPageScenario(pageScenario, page) {
  try {
    if (Array.isArray(pageScenario.url)) {
      pageScenario.url = new Set(pageScenario.url)
    } else if (typeof pageScenario.url === 'string') {
      pageScenario.url = new Set([pageScenario.url])
    } else {
      throw Error('url should be either string or string[]')
    }

    for (const url of pageScenario.url) {
      pageScenario.url.delete(url)
      scannedUrl.add(url)
      await runUrl(url, pageScenario, page)
    }
  } catch (err) {
    logger.error(err)
  }
}

function hrefToUrl(href, baseUrl) {
  if (!href) return ''

  try {
    return (new URL(href, baseUrl)).toString()
  } catch (err) {
    logger.error(err)
    return ''
  }
}

async function getNextUrl(baseUrl, nextUrl, page) {
  logger.debug(`Searching next url`)

  let href = ''

  if (nextUrl.selector) {
    logger.debug(`Searching for: ${nextUrl.selector}`)
    const child = await page.$(nextUrl.selector)
    href = await child.evaluate(node => node.href)
  } else if (nextUrl.xpath) {
    logger.debug(`Searching for: ${nextUrl.xpath}`)
    const [child] = await page.$x(nextUrl.xpath)
    href = await child.evaluate(node => node.href)
  } else {
    throw new Error('Missing selector or xpath')
  }

  return href && hrefToUrl(href, baseUrl)
}

async function runUrl(url, pageScenario, page) {
  logger.info(`Scanning: ${url}`)

  await page.goto(url, {
    waitUntil: 'domcontentloaded',
  })

  if (pageScenario.waitForSelector) {
    await page.waitForSelector(pageScenario.waitForSelector)
  }

  if (pageScenario.waitForXPath) {
    await page.waitForXPath(pageScenario.waitForXPath)
  }

  const output = await runSchema(pageScenario.schema, page)

  if (pageScenario.notifications) {
    for (const channel of pageScenario.notifications) {
      await informPageResult(channel, output)
    }
  }

  if (pageScenario.nextUrl) {
    const nextUrl = await getNextUrl(url, pageScenario.nextUrl, page)

    if (nextUrl) {
      if (scannedUrl.has(nextUrl)) {
        logger.debug(`Next url already scanned. Skipping.`)
      } else {
        logger.debug(`Queueing next url: ${nextUrl}`)
        pageScenario.url.add(nextUrl)
      }
    }
  }

  const interval = pageScenario.interval ? parseDuration(pageScenario.interval, 'ms') : 0
  const hasNextUrl = pageScenario.url.size > 0
  if (interval && hasNextUrl) {
    logger.info(`Cooling down for ${pageScenario.interval}, ${pageScenario.url.size} urls in the queue`)
    await sleep(interval)
  }
}

async function runSchema(schema, handle) {
  switch (schema.type) {
    case 'array': {
      let children = []

      if (schema.selector) {
        logger.debug(`Searching for: ${schema.selector}`)
        children = await handle.$$(schema.selector)
      } else if (schema.xpath) {
        logger.debug(`Searching for: ${schema.xpath}`)
        children = await handle.$x(schema.xpath)
      } else {
        throw new Error('Missing selector or xpath')
      }


      const output = []

      for (const child of children) {
        output.push(await runSchema(schema.items, child))
      }

      return output
    }
    case 'object': {
      const output = {}

      for (const property in schema.properties) {
        output[property] = await runSchema(schema.properties[property], handle)
      }

      return output
    }
    case 'string': {
      if (schema.selector) {
        logger.debug(`Searching for: ${schema.selector}`)
        return await handle.$eval(schema.selector, node => node.textContent)
      } else if (schema.xpath) {
        logger.debug(`Searching for: ${schema.xpath}`)
        const children = await handle.$x(schema.xpath)
        const contents = await Promise.all(children.map(child => child.evaluate(node => node.textContent)))
        return contents.join('\n')
      } else {
        throw new Error('Missing selector or xpath')
      }
    }
    default:
      throw new Error('Unsupported schema type')
  }
}

async function informPageResult(channel, data) {
  if (channel.type === 'webhook') {
    return await sendWebhook(channel.url, data)
  } else if (channel.type === 'console') {
    logger.info(JSON.stringify(data))
    return
  } else {
    throw new Error('Unsupported notification channel')
  }
}